@@ -18,6 +18,7 @@
  */
 
 #include "kio_gdrive.h"
+#include "gdrivebackend.h"
 #include "gdrivedebug.h"
 #include "gdrivehelper.h"
 #include "gdriveurl.h"
@@ -83,6 +84,8 @@ KIOGDrive::KIOGDrive(const QByteArray &protocol, const QByteArray &pool_socket,
 {
     Q_UNUSED(protocol);
 
+    m_accountManager.reset(new AccountManager);
+
     qCDebug(GDRIVE) << "KIO GDrive ready: version" << GDRIVE_VERSION_STRING;
 }
 
@@ -105,7 +108,7 @@ KIOGDrive::Action KIOGDrive::handleError(const KGAPI2::Job &job, const QUrl &url
             return Fail;
         case KGAPI2::Unauthorized: {
             const AccountPtr oldAccount = job.account();
-            const AccountPtr account = m_accountManager.refreshAccount(oldAccount);
+            const AccountPtr account = m_accountManager->refreshAccount(oldAccount);
             if (!account) {
                 error(KIO::ERR_CANNOT_LOGIN, url.toDisplayString());
                 return Fail;
@@ -132,6 +135,47 @@ KIOGDrive::Action KIOGDrive::handleError(const KGAPI2::Job &job, const QUrl &url
     return Fail;
 }
 
+void KIOGDrive::fileSystemFreeSpace(const QUrl &url)
+{
+    const auto gdriveUrl = GDriveUrl(url);
+    const QString accountId = gdriveUrl.account();
+    if (accountId == QLatin1String("new-account")) {
+        finished();
+        return;
+    }
+    if (!gdriveUrl.isRoot()) {
+        AboutFetchJob aboutFetch(getAccount(accountId));
+        if (runJob(aboutFetch, url, accountId)) {
+            const AboutPtr about = aboutFetch.aboutData();
+            if (about) {
+                setMetaData(QStringLiteral("total"), QString::number(about->quotaBytesTotal()));
+                setMetaData(QStringLiteral("available"), QString::number(about->quotaBytesTotal() - about->quotaBytesUsedAggregate()));
+                finished();
+                return;
+            }
+        }
+    }
+    error(KIO::ERR_CANNOT_STAT, url.toDisplayString());
+}
+
+AccountPtr KIOGDrive::getAccount(const QString &accountName)
+{
+    return m_accountManager->account(accountName);
+}
+
+void KIOGDrive::virtual_hook(int id, void *data)
+{
+    switch (id) {
+        case SlaveBase::GetFileSystemFreeSpace: {
+            QUrl *url = static_cast<QUrl *>(data);
+            fileSystemFreeSpace(*url);
+            break;
+        }
+        default:
+            SlaveBase::virtual_hook(id, data);
+    }
+}
+
 KIO::UDSEntry KIOGDrive::fileToUDSEntry(const FilePtr &origFile, const QString &path) const
 {
     KIO::UDSEntry entry;
@@ -182,29 +226,50 @@ void KIOGDrive::openConnection()
     qCDebug(GDRIVE) << "Ready to talk to GDrive";
 }
 
+KIO::UDSEntry KIOGDrive::accountToUDSEntry(const QString &accountNAme)
+{
+    KIO::UDSEntry entry;
 
+    entry.insert(KIO::UDSEntry::UDS_NAME, accountNAme);
+    entry.insert(KIO::UDSEntry::UDS_DISPLAY_NAME, accountNAme);
+    entry.insert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
+    entry.insert(KIO::UDSEntry::UDS_SIZE, 0);
+    entry.insert(KIO::UDSEntry::UDS_ACCESS, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
+    entry.insert(KIO::UDSEntry::UDS_ICON_NAME, QStringLiteral("folder-gdrive"));
 
-
-
-
+    return entry;
+}
 
 void KIOGDrive::createAccount()
 {
-    const KGAPI2::AccountPtr account = m_accountManager.account(QString());
-    redirection(QUrl(QStringLiteral("gdrive:/%1").arg(account->accountName())));
+    const KGAPI2::AccountPtr account = m_accountManager->createAccount();
+    if (!account->accountName().isEmpty()) {
+        // Redirect to the account we just created.
+        redirection(QUrl(QStringLiteral("gdrive:/%1").arg(account->accountName())));
+        finished();
+        return;
+    }
+
+    if (m_accountManager->accounts().isEmpty()) {
+        error(KIO::ERR_SLAVE_DEFINED, i18n("There are no Google Drive accounts enabled. Please add at least one."));
+        return;
+    }
+
+    // Redirect to the root, we already have some account.
+    redirection(QUrl(QStringLiteral("gdrive:/")));
     finished();
 }
 
 void KIOGDrive::listAccounts()
 {
-    const auto accounts = m_accountManager.accounts();
+    const auto accounts = m_accountManager->accounts();
     if (accounts.isEmpty()) {
         createAccount();
         return;
     }
 
     Q_FOREACH (const QString &account, accounts) {
-        const KIO::UDSEntry entry = AccountManager::accountToUDSEntry(account);
+        const KIO::UDSEntry entry = accountToUDSEntry(account);
         listEntry(entry);
     }
     KIO::UDSEntry newAccountEntry;
@@ -460,7 +525,7 @@ void KIOGDrive::stat(const QUrl &url)
         return;
     }
     if (gdriveUrl.isAccountRoot()) {
-        const KIO::UDSEntry entry = AccountManager::accountToUDSEntry(accountId);
+        const KIO::UDSEntry entry = accountToUDSEntry(accountId);
         statEntry(entry);
         finished();
         return;
@@ -840,12 +905,12 @@ void KIOGDrive::del(const QUrl &url, bool isfile)
 
     // If user tries to delete the account folder, remove the account from the keychain
     if (gdriveUrl.isAccountRoot()) {
-        const KGAPI2::AccountPtr account = m_accountManager.account(accountId);
-        if (!account) {
+        const KGAPI2::AccountPtr account = getAccount(accountId);
+        if (account->accountName().isEmpty()) {
             error(KIO::ERR_DOES_NOT_EXIST, accountId);
             return;
         }
-        m_accountManager.removeAccount(accountId);
+        m_accountManager->removeAccount(accountId);
         finished();
         return;
     }
